if __package__=="eznet_keras.models":
    from ..utils import *
    from .keras_smart_module import *
    from .dense_block import *
else:
    import os, sys
    current_dir = os.path.dirname(os.path.abspath(__file__))
    parent_dir = os.path.dirname(current_dir)
    sys.path.append(parent_dir)
    sys.path.append(current_dir)
    from utils import *
    from keras_smart_module import *
    from dense_block import *

import warnings

class Recurrent_Network(KerasSmartModel):
    
    sample_hparams = {
        # General and I/O parameters
        'model_name': 'Recurrent_Network',
        'in_features': 10,
        'out_features': 3,
        'sequence_length': 13,
        'final_rnn_return_sequences': False,
        'flatten_after_rnn': False,
        'permute_output': False,
        # RNN layer parameters
        'rnn_type': 'LSTM',
        'rnn_hidden_size': 20,
        'rnn_bidirectional': False,
        'rnn_depth': 2,
        'rnn_input_dropout': None,
        'rnn_recurrent_dropout': None,
        'rnn_layer_dropout': 0.1,
        'rnn_params': None,
        # Dense layer parameters
        'include_dense_layers': True,
        'dense_width': 30,
        'dense_depth': 2,
        'dense_dropout': 0.2,
        'dense_activation': 'relu',
        'dense_activation_params': None,
        'norm_layer_type': 'BatchNormalization',
        'norm_layer_params': None,
        'norm_layer_position': 'before',
        # Output layer parameters
        'include_output_layer': True,
        'output_activation': "softmax",
        'output_activation_params': None,
        # Training procedure parameters
        'l2_reg': 0.0001,
        'batch_size': 16,
        'epochs': 2,
        'validation_data': [0.05,'testset'],
        'early_stopping_patience_epochs': 10,
        'learning_rate': 0.001,
        'exponential_decay_rate': 0.95,
        'loss_function': 'categorical_crossentropy',
        'metrics':['accuracy'],
        'optimizer': 'Adam',
        'optimizer_params': None,
        'checkpoint_path':None,
        'early_stopping_monitor':'loss',
        'early_stopping_mode':'min',
        'early_stopping_value':1.0e-6
    }
    
    def __init__(self, hparams:dict=None):
        """Sequence to Dense network with RNN for time-series classification, regression, and forecasting, as well as NLP applications.
        This network uses any RNN layers as encoders to extract information from input sequences, and (optionally) fully-connected 
        multilayer perceptrons (Dense) to decode the sequence into an output, which can be class probabilitites 
        (timeseries classification), a continuous number (regression), or an unfolded sequence (forecasting) of a 
        target timeseries. Multiple stacked RNN layers are supported, as well as bidirectional RNNs. Furthermore, the dense blocks in the decoder part are by default assumed to
        contain a dense layer, followed (optionally) by a normalization, activation, and dropout layer. Multiple activation functions, loss functions, optimizers, etc. are
        supported, meaning this class can be used to construct virtually any serial RNN network possible, with or without dense layers, with or without output layers.
        In this class, parallelization or skip connections are not yet supported.
        The RNN section supports input dropout, recurrent dropout, and layer dropout (dropout between two stacked layers of RNN, for example).

        ### Usage

        `net = Recurrent_Network(hparams)` where `hparams` is dictionary of hyperparameters containing the keys described below.
        
        - Inspect the `sample_hparams` class attribute for a template of the hyperparameters dictionary.
        - Many keys, especially those that have "(list of)" in their description, can be either scalar items or lists of items. If they are scalar, they will be broadcasted to all 
          RNN/Dense blocks. If they are lists, they must have the same length as the number (depth) of RNN/Dense blocks, i.e. `rnn_depth`/`dense_depth` hyperparameters.
        - The Training Procedure section of the hyperparameters are optional, and will only be used by the parent class `KerasSmartModel` if the model is trained using its methods.
          Otherwise, it is not necessary at all for building a network.
        - Activation functions come from `tf.keras.activations` and need to be wrapped in a `tf.keras.layers.Activation` layer, and do not accept any kwargs. Activation layers
          come from `tf.keras.layers`, can accept kwargs in their constructors, and do not need to be wrapped in an `Activation` layer. Activation functions all have lower-case
          names, but activation layers are classes and have every-word-capitalized names. We will use this as a clue to recognize what the user wants, and perform accordingly.
          Most activation layers also have correpsonding activation functions (like `tf.keras.activations.relu` function and `tf.keras.layers.ReLU` layer). 
          However, some activations are only available as functions (such as `tf.keras.activations.sigmoid`), and some are only available as layers 
          (such as `tf.keras.layers.LeakyReLU`). Choose accordingly.
        
        The keys of the hyperparameter dictionary are as follows:
        
        #### General and I/O hyperparameters
        
            - `model_name` (str): Name of the model, can be used later for saving, etc.
            - `in_features` (int): Number of features of the input.
            - `out_features` (int): Number of features of the output, assuming there is an output layer (including a dense layer and an activation function)
            - `sequence_length` (int): Length of the sequence.
            - `final_rnn_return_sequences` (bool): Whether the final RNN returns a whole sequence rather than its last time step. Defaults to False.
                This is useful if a sequence-to-sequence architecture is intended, or all information of all time steps of the final RNN layer is desired.
            - `flatten_after_rnn` (bool): If `final_rnn_return_sequences` is True, whether to flatten the final RNN output sequence into a single feature vector.
                This way, all information of all time steps of the final RNN layer will be preserved, and the 3D output of the final RNN layer will be flattened to a 2D tensor.
                If `final_rnn_return_sequences` is False, this parameter is ignored. Defaults to False.
            - `permute_output` (bool): If `'final_rnn_return_sequences' = True` and `'flatten_after_rnn' = False`, whether to permute the output sequence to be 
                `(batch, features, sequence_length)`. Default is False.
            
            
        #### RNN architecture hyperparameters
        
            - `rnn_type` (str|list): (list of) RNN types, options are "LSTM", "GRU", "SimpleRNN", etc. They should be class names or custom RNN layer classes (not instances).
            - `rnn_hidden_size` (int|list): (list of) RNN layer hidden sizes. Default is 16.
            - `rnn_bidirectional` (bool): Whether the RNN layers are bidirectional or not. Default is False.
            - `rnn_depth` (int): Number of stacked RNN layers. Default is 1.
            - `rnn_input_dropout` (float|list): (list of) Dropout rates, if any, of the RNN layer inputs. 
                Please note that using dropout in RNN layers is generally discouraged, for it decreases determinism during inference.
            - `rnn_recurrent_dropout` (float|list): (list of) Dropout rates, if any, of the RNN layer recurrent states.
            - `rnn_layer_dropout` (float|list): (list of) Dropout rates, if any, of the RNN layer outputs. This will be applied to all time steps equally.
            - `rnn_params` (dict): A single dictionary of kwargs for the RNN layers' constructors. Default is None.
              If specified, the keys in this dictionary will not only add to, but also overwrite any existing arguments this class passes to the RNN layer constructor.
            
            
        #### Dense network architecture hyperparameters
        
            - `include_dense_layers` (bool): Whether to include a Dense network after the RNN layers. Default is True.
            - `dense_width` (int|list): (list of) Widths of the Dense network. It can be a number (for all) or a list holding width of each hidden layer.
            - `dense_depth` (int): Depth (number of hidden layers) of the Dense network, not including the output layer.
            - `dense_activation` (str|list): (list of) Activation functions for hidden layers of the Dense network.
                It can be activation function ("relu", "sigmoid", "softmax", etc.), activation layer ("ReLU", "LeakyReLU", "Softmax", etc.), or a custom Keras layer class 
                (not instance)
            - `dense_activation_params` (dict|list): (list of) Dictionaries of parameters for the activation function constructors of the Dense network.
                Ignored if the activation function is a lower-case function name. By the way, for `LeakyReLU`, the left-hand slope is specified by the `alpha` key.
            - `norm_layer_type` (str|list): (list of) Types of normalization layers to use in the dense section, if any. 
                Options are "BatchNormalization", "LayerNormalization", etc. It can be a layer name or a Keras Layer class (not instance).
            - `norm_layer_params` (dict|list): (list of) Dictionaries of parameters for the normalization layer constructors.
            - `norm_layer_position` (str|list): (list of) Whether the normalization layer should come 'before' or 'after' the activation of each hidden layer in the dense network.
            - `dense_dropout` (float|list): (list of) Dropout rates (if any) for the hidden layers of the Dense network.

        #### Output layer hyperparameters
        
            - `include_output_layer` (bool): Whether to include an output layer, regardless of whether there was a Dense network. Default is True.
                The output layer is a dense layer followed optionally by an activation.
            - `output_activation` (str): Activation function for the output layer of the Dense network, if any.
                Like the `dense_activation` key, it can be activation function name, layer name, or a custom Keras Layer class (not instance).
                **NOTE** If the `loss_function` is `sparse_categorical_crossentropy`, then no output activation is erquired.
                However, if it is `categorical_crossentropy` (a.k.a. negative log-likelihood), then you must specify an output activation as in "softmax".
            - `output_activation_params` (dict): Dictionary of parameters for the activation function constructor of the output layer.
            
        #### Training procedure hyperparameters
        
            - `batch_size` (int): Minibatch size, the expected input size of the network. Defaults to 32.
            - `learning_rate` (float): Initial learning rate of training. Defaults to 0.001. Will be given directly to the optimization function.
            - `exponential_decay_rate` (float): Exponential decay rate for learning rate, if any.
            - `optimizer` (str): Optimizer. Examples: 'Adam', 'SGD', 'RMSProp', etc. It can be the name of a Keras optimizer or custom optimizer class. Will be given directly
                to the `compile` function. If it is custom, it should be a class, not an instance. Constructor kwargs should be given via the next key.
            - `optimizer_params` (dict): Additional parameters of the optimizer constructor, if any. Defaults to None.
            - `epochs` (int): Maximum number of epochs for training. Defaults to 2.
            - `early_stopping_patience_epochs` (int): Epochs to tolerate unimproved (val) loss, before early stopping. Defaults to None.
            - `validation_data` (list): Portion of validation data. Should be a tuple like [validation split, dataset as in 'trainset' or 'testset']. Defaults to None.
            - `l2_reg` (float): L2 regularization parameter. Defaults to None.
            - `loss_function` (str): Loss function. Examples: 'binary_crossentropy', 'categorical_crossentropy', 'mse', etc. Will be given directly to the `compile` function.
            - `metrics` (list): list of metrics for Keras compilation, such as ['accuracy']. Will be given directly to the `compile` function.
            - `checkpoint_path` (str): Path to the directory where checkpoints will be saved at every epoch. The path does not need to exist beforehand.
            - `early_stopping_monitor` (str): Monitor whose critical value will cause early stopping. Default is 'loss', but 'val_loss' is typically used.
            - `early_stopping_mode` (str): Mode of the parameter whose critical value will be used for early stopping. Deafults to 'min' for any error. 'max' is for accuracy, etc.
            - `early_stopping_value` (float): Value of the monitor at which point training will stop becasue the critical value has been reached.

        ### Returns
        
        Returns a `tf.keras.models.Model` object that can be trained and used accordingly.
        Run `net.summary()` afterwards to see what you have inside the network.
        The returned model is an instance of `KerasSmartModel`, which has built-in functions for training, evaluation, etc.
        """
        super(Recurrent_Network, self).__init__(hparams)
        hparams = hparams if hparams is not None else self.sample_hparams
        
        # Read General and I/O hyperparameters
        self._model_name = hparams.get("model_name") if "model_name" in hparams else "Recurrent_Network"
        self._in_features = hparams["in_features"]
        self._out_features = hparams.get("out_features")
        self._L = int(hparams["sequence_length"])
        self._final_rnn_return_sequences = True if hparams.get("final_rnn_return_sequences") else False
        self._flatten_after_rnn = True if hparams.get("flatten_after_rnn") else False
        if self._flatten_after_rnn:
            assert self._final_rnn_return_sequences, "If `flatten_after_rnn` is True, then `final_rnn_return_sequences` must be True as well. Check your hyperparameters."
        self._permute_output = True if hparams.get("permute_output") else False
        if self._permute_output:
            assert self._final_rnn_return_sequences and not self._flatten_after_rnn, \
                "If `permute_output` is True, then `final_rnn_return_sequences` must be True and `flatten_after_rnn` must be False. Check your hyperparameters."
            
        # Read RNN hyperparameters
        self._rnn_type = hparams["rnn_type"]
        self._rnn_params = hparams.get("rnn_params")
        self._rnn = getattr(tf.keras.layers, self._rnn_type) if isinstance(self._rnn_type, str) else self._rnn_type
        self._rnn_hidden_size = hparams["rnn_hidden_size"] if hparams.get("rnn_hidden_size") else 16
        self._rnn_depth = hparams["rnn_depth"] if hparams.get("rnn_depth") else 1
        self._bidirectional = True if hparams.get("rnn_bidirectional") else False
        self._rnn_input_dropout = hparams.get("rnn_input_dropout")
        self._rnn_recurrent_dropout = hparams.get("rnn_recurrent_dropout")
        self._rnn_layer_dropout = hparams.get("rnn_layer_dropout")        
            
        # Read Dense hyperparameters
        self._include_dense_layers = hparams.get("include_dense_layers") if "include_dense_layers" in hparams else \
            (hparams.get('dense_depth') is not None and hparams.get('dense_depth') > 0)
        self._dense_width = hparams["dense_width"]
        self._dense_depth = hparams["dense_depth"] if hparams.get("dense_depth") else 0
        self._dense_activation = hparams.get("dense_activation")
        self._dense_activation_params = hparams.get("dense_activation_params")
        self._norm_layer_type = hparams.get("norm_layer_type")
        self._norm_layer_params = hparams.get("norm_layer_params")
        self._norm_layer_position = hparams.get("norm_layer_position")
        self._dense_dropout = hparams.get("dense_dropout")
        
        # Read Output hyperparameters
        self._include_output_layer = hparams.get("include_output_layer") if "include_output_layer" in hparams else True
        if not self._include_output_layer:
            assert self._out_features is None, "If `include_output_layer` is False, then `out_features` must be None. Check your hyperparameters."
        self._output_activation = hparams.get("output_activation")
        self._output_activation_params = hparams.get("output_activation_params")
        
        # Calculate shape-related variables
        self._N = int(self._batch_size)
        self._D = int(2 if self._bidirectional else 1)
        self._H_in = int(self._in_features)
        self._H_cell = self._rnn_hidden_size
        self._H_out = self._H_cell # Projection is not supported in Keras, unlike PyTorch.
        
        # Calculate batch input shape
        self.batch_input_shape = (self._N, self._L, self._H_in)
        
        # Calculate width (number of features) of the output layer (final layer, whatever that is)
        if self._include_output_layer:
            self._output_width = self._out_features
        elif self._include_dense_layers:
            self._output_width = self._dense_width if isinstance(self._dense_width, int) else self._dense_width[-1]
        else:
            self._output_width = (self._L if self._flatten_after_rnn else 1) * self._D * self._H_out
        
        # Calculate batch output shape
        self._output_is_sequence = (self._final_rnn_return_sequences and not self._flatten_after_rnn)
        if self._output_is_sequence:
            self.batch_output_shape = (self._N, self._output_width, self._L) if self._permute_output else (self._N, self._L, self._output_width)
        else:
            self.batch_output_shape = (self._N, self._output_width)
        
        # Initializing sequential network
        self.net = tf.keras.models.Sequential(name=self._model_name)
        self.net.add(tf.keras.Input((self._L, self._H_in)))
        
        # Generate vectors out of scalar hyperparameters
        vectorify = lambda x: x if isinstance(x, (list, tuple)) else [x] * self._rnn_depth
        self._H_cell_vec = vectorify(self._H_cell)
        self._rnn_input_dropout_vec = vectorify(self._rnn_input_dropout)
        self._rnn_recurrent_dropout_vec = vectorify(self._rnn_recurrent_dropout)
        self._rnn_layer_dropout_vec = vectorify(self._rnn_layer_dropout)
        
        # Construct RNN layers
        for i in range(self._rnn_depth):
            _kwargs = {
                'units': self._H_cell_vec[i],
                'dropout': self._rnn_input_dropout_vec[i] if self._rnn_input_dropout_vec[i] is not None else 0.0,
                'recurrent_dropout': self._rnn_recurrent_dropout_vec[i] if self._rnn_recurrent_dropout_vec[i] is not None else 0.0,
                'return_sequences':(True if self._rnn_depth > 1 else self._final_rnn_return_sequences), 
                'kernel_regularizer':(tf.keras.regularizers.L2(self._l2_reg) if self._l2_reg else None)
            }
            if self._rnn_params is not None: 
                _kwargs.update(self._rnn_params)
            if i != self._rnn_depth-1:
                _kwargs.update({'return_sequences':True})
            else:
                _kwargs.update({'return_sequences':self._final_rnn_return_sequences})
            self.net.add(tf.keras.layers.Bidirectional(self._rnn(**_kwargs)) if self._bidirectional else self._rnn(**_kwargs))
            if self._rnn_layer_dropout_vec[i] is not None:
                self.net.add(tf.keras.layers.Dropout(self._rnn_layer_dropout_vec[i], 
                    noise_shape=((self._N, 1, self._H_cell_vec[i] * self._D) if self._final_rnn_return_sequences else None)))
        
        # Flatten the output sequence before decoding if necessary
        if self._flatten_after_rnn:
            self.net.add(tf.keras.layers.Flatten())
        
        # Construct Dense Network
        if self._include_dense_layers and self._dense_depth > 0:
                
            # Generate arrays containing parameters of each Dense Block (Every block contains a linear, normalization, activation, and dropout layer).
            self._dense_width_vec = self._gen_hparam_vec_for_dense(self._dense_width, 'dense_width')
            self._dense_activation_vec = self._gen_hparam_vec_for_dense(self._dense_activation, 'dense_activation')
            self._dense_activation_params_vec = self._gen_hparam_vec_for_dense(self._dense_activation_params, 'dense_activation_params')
            self._dense_norm_layer_type_vec = self._gen_hparam_vec_for_dense(self._norm_layer_type, 'norm_layer_type')
            self._dense_norm_layer_params_vec = self._gen_hparam_vec_for_dense(self._norm_layer_params, 'norm_layer_params')
            self._dense_norm_layer_position_vec = self._gen_hparam_vec_for_dense(self._norm_layer_position, 'norm_layer_position')
            self._dense_dropout_vec = self._gen_hparam_vec_for_dense(self._dense_dropout, 'dense_dropout')
            
            # Construct the dense layers
            # in_size = self._dense_input_size
            for i in range(self._dense_depth):
                out_size = self._dense_width_vec[i]
                add_dense_block(self.net, output_size=out_size, input_shape=None, activation=self._dense_activation_vec[i], activation_params=self._dense_activation_params_vec[i], 
                        norm_layer_type=self._dense_norm_layer_type_vec[i], norm_layer_position=self._dense_norm_layer_position_vec[i], 
                        norm_layer_params=self._dense_norm_layer_params_vec[i], dropout=self._dense_dropout_vec[i], 
                        kernel_regularizer=(tf.keras.regularizers.L2(self._l2_reg) if self._l2_reg else None))
                # in_size = out_size
        
        # Output layer
        if self._include_output_layer:
            self.net.add(tf.keras.layers.Dense(self._output_width, kernel_regularizer=(tf.keras.regularizers.L2(self._l2_reg) if self._l2_reg else None)))
            if self._output_activation:
                if isinstance(self._output_activation, str):
                    if self._output_activation.lower() == self._output_activation:
                        self.net.add(tf.keras.layers.Activation(self._output_activation))
                    elif self._output_activation_params:
                        self.net.add(getattr(tf.keras.layers, self._output_activation)(**self._output_activation_params))
                    else:
                        self.net.add(getattr(tf.keras.layers, self._output_activation)())
                elif self._output_activation_params:
                    self.net.add(self._output_activation(**self._output_activation_params))
                else:
                    self.net.add(self._output_activation())
            
        # Permute if necessary
        if self._permute_output:
            self.net.add(tf.keras.layers.Permute((2,1)))
            
            
        
    def _gen_hparam_vec_for_dense(self, hparam, hparam_name, **kwargs):
        return generate_array_for_hparam(hparam, self._dense_depth, hparam_name=hparam_name, count_if_not_list_name='dense_depth', **kwargs)
    
    def call(self, x, *args, **kwargs):
        return self.net(x, *args, **kwargs)



if __name__ == '__main__':
    
    # Everything default
    # test_keras_model_class(Recurrent_Network, save_and_export=False)
    # SUCCESS!
    
    # No dense or output, just an RNN
    # hparams = Recurrent_Network.sample_hparams.copy()
    # hparams.update({
    #     'include_dense_layers': False,
    #     'include_output_layer': False,
    #     'out_features': None,
    #     'final_rnn_return_sequences': True
    # })
    # test_keras_model_class(Recurrent_Network, hparams=hparams, save_and_export=True)
    # SUCCESS!
    
    # RNN and Dense, no output
    # hparams = Recurrent_Network.sample_hparams.copy()
    # hparams.update({
    #     'include_output_layer': False,
    #     'out_features': None,
    #     'final_rnn_return_sequences': True, #False
    #     'permute_output': True #False
    # })
    # test_keras_model_class(Recurrent_Network, hparams=hparams, save_and_export=False)
    # SUCCESS!
    
    # Include everything, but have flattened sequences
    # hparams = Recurrent_Network.sample_hparams.copy()
    # hparams.update({
    #     'final_rnn_return_sequences': True, #False
    #     'flatten_after_rnn': True, #False
    #     'permute_output': False #True
    # })
    # test_keras_model_class(Recurrent_Network, hparams=hparams, save_and_export=False)
    # SUCCESS!
    
    pass
    
    
    


import numpy as np
from keras import Sequential
from keras.layers import Dense

#create random data
test_x = np.random.rand(10, 10).astype('f')
test_y = np.random.rand(10).astype('f')
model = Sequential([
    Dense(1, input_dim=10)
])
model.compile(loss='mse', optimizer='adam')

#train model by 1 iteration
model.fit(test_x, test_y, epochs=1, verbose=False)

#predict
data = np.array([[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]])
prediction = model.predict(data)
print(prediction)

#save model
from keras2cpp import export_model
export_model(model, 'example.model')



if __name__=="eznet_keras.test":
    from .utils import *
    from .models import *
else:
    from utils import *
    from models import *

def testcalc_image_size():
    size = calc_image_size(size_in=[28,28,28], kernel_size=3, padding='valid', stride=2, dilation=1)
    print(size)

def testgenerate_geometric_array():
    array = generate_geometric_array(init=256, count=4, direction='up', powers_of_two=True)
    print(array)
    
def testgenerate_array_for_hparam():
    array = generate_array_for_hparam(
        [1,2], count_if_not_list=4, 
        hparam_name='parameter', count_if_not_list_name='its count',
        check_auto=True, init_for_auto=2, powers_of_two_if_auto=True,
        direction_if_auto='up')
    print(array)

def test_dense_block():
    dense_block = Dense_Block([256], 128, 'relu', None, 'BatchNormalization', 'before', None, 0.1, tf.keras.regularizers.L2(0.0001))
    print(dense_block)
    print(dense_block.summary())
    x = np.random.randn(32,256)
    y = dense_block(x)
    print("Input shape:  ", x.shape)
    print("Output shape: ", y.shape)

def test_conv_block():
    conv_block = Conv_Block([28,28,3], 32, conv_dim=2, input_image=[28,28], conv_kernel_size=3, conv_padding='valid', conv_stride=1, conv_dilation=1, conv_params=None, 
                            conv_activation='relu', conv_activation_params=None, norm_layer_position='before', norm_layer_type='BatchNormalization', norm_layer_params=None, 
                            pool_type='Max', pool_kernel_size=2, pool_padding='valid', pool_stride=1, pool_params=None, dropout=0.1, min_image_dim=8,
                            kernel_regularizer=tf.keras.regularizers.L2(0.0001))
    print(conv_block)
    print(conv_block.summary())
    x = np.random.randn(32,28,28,3)
    y = conv_block(x)
    print("Input shape:  ", x.shape)
    print("Output shape: ", y.shape)
    
def test_conv_network():
    test_keras_model_class(Conv_Network)

def test_recurrent_network():
    test_keras_model_class(Recurrent_Network)
    
def test_ann_network():
    test_keras_model_class(ANN)




if __name__ == '__main__':
    
    # testcalc_image_size()
    # testgenerate_geometric_array()
    # testgenerate_array_for_hparam()
    
    # test_dense_block()
    # test_conv_block()
    
    # test_conv_network()
    # test_recurrent_network()
    # test_ann_network()
    
    
    pass


if __package__=="eznet_keras.models":
    from .keras_smart_module import *
    from .dense_block import *
else:
    import sys, os
    sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
    from keras_smart_module import *
    from dense_block import *

import warnings

class ANN(KerasSmartModel):
    
    sample_hparams = {
        # General and I/O parameters
        "model_name": "ANN",
        "input_shape": [10],
        "output_size": 3,
        # Architecture parameters
        "width": 32,
        "depth": 2,
        "hidden_activation": "LeakyReLU",
        "hidden_activation_params": {'alpha': 0.1},
        "norm_layer_type":"BatchNormalization",
        "norm_layer_position": "before",
        "norm_layer_params": None,
        "dropout": 0.2,
        "include_output_layer": True,
        "output_activation": "Softmax",
        "output_activation_params": None,
        # Training procedure parameters
        "learning_rate": 0.001,
        "exponential_decay_rate": 0.9,
        "batch_size": 32,
        "epochs": 2,
        "early_stopping_patience_epochs": 2,
        "validation_data": [0.05,'trainset'],
        "l2_reg": 0.0001,
        "loss_function": "categorical_crossentropy",
        'optimizer': 'Adam',
        'optimizer_params': None,
        'metrics': ['accuracy'],
        'checkpoint_path': None,
        'early_stopping_monitor': 'loss',
        'early_stopping_mode': 'min',
        'early_stopping_value': 1.0e-6
    }
    
    
    def __init__(self, hparams:dict=None):
        """Typical Artificial Neural Network class, also known as multilayer perceptron. This class will create a fully connected feedforward artificial neural network.
        It can be used for classification, regression, etc. It basically encompasses enough options to build all kinds of serial ANNs with any number of 
        inputs, outputs, layers with custom or arbitrary width or depth, etc. Supports multiple activation functions for hidden layers and the output layer. The netwrok consists
        of many `Dense_Block` layers, each consisting of a linear (dense) layer, followed optionally by a normalization, activation, and dropout layers. There is optionally an 
        output layer after all the `Dense_Block` instances. This class can be used for making virtually any kind of serial MLP. No parallelism or skip connections are supported
        in this calss yet.
        
        ### Usage
        
        `net = ANN(hparams)` where `hparams` is the dictionary of hyperparameters.
        
        - Inspect the `sample_hparams` class attribute to find a complete sample of hyperparameters dictionary.
        - The training procedure section of the hyperparameters is the same as in `KerasSmartModel` class, and is only used in training. Otherwise, it is not necessary for building
          the model itself.
        - Many hyperparameters, especially those that have a (list of) in their description below, can be either None, in which case that item will not be included, or a scalar
          value which will be broadcasted to all dense blocks, or a list with the same length as the `depth` hyperparameter, in which case each item will be used for the 
          corresponding dense block. Use None for any dense block for which an item or hyperparameter should not be included or applied.
        - Every `Dense_Block` by default is assumed to have a dense layer, followed by a normalization, activation, and dropout layers. We also assume that there will be an output
          layer at the end, unless `include_output_layer` hyperparameter is set to `False`. Normalization layer can come `before` or `after` the activation.
        - Activation functions come from `tf.keras.activations` and need to be wrapped in a `tf.keras.layers.Activation` layer, and do not accept any kwargs. Activation layers
          come from `tf.keras.layers`, can accept kwargs in their constructors, and do not need to be wrapped in an `Activation` layer. Activation functions all have lower-case
          names, but activation layers are classes and have every-word-capitalized names. We will use this as a clue to recognize what the user wants, and perform accordingly.
          Most activation layers also have correpsonding activation functions (like `tf.keras.activations.relu` function and `tf.keras.layers.ReLU` layer). 
          However, some activations are only available as functions (such as `tf.keras.activations.sigmoid`), and some are only available as layers 
          (such as `tf.keras.layers.LeakyReLU`). Choose accordingly.

        The hyperparameters dictionary should include the following keys:
        
            - `input_shape` (list): shape of the input layer minus the batch size. It can have any number of elements. In higher dimensionalities the dense layer will be applied to
                all feature maps, such that only the last dimension will change. An input of `[batch, dim1, dim2, in_dim]` will come out as `[batch, dim1, dim2, layer_size]`
                from a dense block.
            - `output_size` (int): number of outputs to predict, i.e. size of the output layer, if there is going to be an output layer.
            - `width` (int|list): (list of) hidden layer widths. 
                a number sets them all the same, and a list/array sets each hidden layer according to the list.
            - `depth` (int): Specifies the depth of the network (number of hidden layers).
                It must be specified unless `width` is provided as a list. Then the depth will be inferred form it.
            - `hidden_activation` (str): (list of) Activations. It can be an activation function name ("relu","sigmoid","tanh", etc.), an activation layer name ("ReLU", 
                "LeakyReLU", "Softmax", etc.), or a custom Keras Layer class (not instance). Defaults to None.
            - `hidden_activation_params` (dict): (list of) kwargs parameters for the hidden layer activation constructor, if any. Ignored for any dense block if the activation is
                provided as a lower-case function name. By the way, the slope of the negative section in `LeakyReLU` is `alpha`. Defaults to None.
            - `norm_layer_type` (str): (list of) Types of normalization layers to use for each hidden layer. Options are "BatchNormalization", "LayerNormalization", etc.
                This can also be a Keras Layer class (not instance) rather than a string. Defaults to None.
            - `norm_layer_position` (str): (list of) where the normalization layer should be included relative to the activation function (if any), 'before' or 'after'.
            - `norm_layer_params` (dict): (list of) Dictionaries of kwargs parameters for the normalization layers' constructors. Defaults to None.
            - `dropout` (float): (list of) the dropout rates after every hidden layer. It should be a probability value between 0 and 1, or None by default.
            - `include_output_layer` (bool): Whether to include an output layer at the end of the network. Defaults to True.
            - `output_activation` (str): Activation of the output layer, if any. Defaults to None.
                For classification problems, you may want to choose "sigmoid" or "softmax".
                That being said, you usually don't need to specify an activation for the output layer at all, if e.g. 'from_logits' is used.
                For regression problems, no activation is needed. It is by default linear, unless you want to manually specify an activation.
            - `output_activation_params` (dict): Parameters for the output activation function, if any. Ignored if lower-case function name is provided for activation.
                Defaults to None.
            - `learning_rate` (float): Initial learning rate of training.
            - `exponential_decay_rate` (float): Exponential decay rate for learning rate, if any. Defaults to None.
            - `optimizer` (str): Optimizer. Examples: "Adam", "SGD" ,"RMSProp", etc. The name of any optimizer class in `tf.keras.optimizers` can be used, or a custom class.
            - `optimizer_params` (dict): Additional parameters of the optimizer, if any. Defaults to None.
            - `batch_size` (int): Minibatch size for training.
            - `epochs` (int): Maximum number of epochs for training.
            - `early_stopping_patience_epochs` (int): Epochs to tolerate unimproved (val) loss, before early stopping.
            - `validation_data` (list): List of [validation_split, 'trainset'|'testset']. Defaults to None.
            - `l2_reg` (float): L2 regularization parameter. Defaults to None.
            - `loss_function` (str): Loss function. Examples: "mse","binary_crossentropy", "categorical_crossentropy", etc. It can also be a loss function instance.
            - `metrics` (list): list of metrics for Keras compilation, e.g. ['accuracy'].
            - `checkpoint_path` (str): Path to the directory where checkpoints will be saved at every epoch.
            - `early_stopping_monitor` (str): Monitor whose critical value will cause early stopping. Default is 'loss', but 'val_loss' is typically used.
            - `early_stopping_mode` (str): Mode of the parameter whose critical value will be used for early stopping. Deafults to 'min' for any error. 'max' is for accuracy, etc.
            - `early_stopping_value` (float): Value of the monitor at which point training will stop becasue the critical value has been reached.
        
        Note that for all such hyperparameters that have a (list of) at the beginning, the entry can be a single item repeated for all hidden layers, or it can be a list of items
        for all hidden layers. If a list is provided, it must have the same length as the depth of the network. 
        
        Also note that depth does not include the input and output layers.
        This gives you the ability to specify different width, dropout rate, normalization layer and its parameters, and so forth.

        ### Returns
        
        It returns a `tf.keras.models.Model` object that corresponds with an ANN model.
        run `net.summary()` afterwards to see what the ANN holds.
        The returned module is a `KerasSmartModel` instance, which is a subclass. It has built-in functions for training, evaluation, etc.
        """
        super(ANN, self).__init__(hparams)
        # Read and store hyperparameters
        if not hparams: hparams=self.sample_hparams
        self._model_name = hparams.get("model_name") if "model_name" in hparams else "ANN"
        self._input_shape = hparams.get("input_shape")
        if self._input_shape is not None:
            assert isinstance(self._input_shape, (list,tuple)), "input_shape must be a list or tuple."
        self._output_size = hparams.get("output_size")
        self._dropout = hparams.get("dropout")
        self._width = hparams.get("width")
        self._depth = hparams.get("depth")
        if self._depth is None:
            assert isinstance(self._width, (list,tuple)), "If depth is not provided, width must be a list or tuple."
            self._depth = len(self._width)
        elif self._width is None:
            assert isinstance(self._depth, int), "If width is not provided, depth must be an integer."
            warnings.warn("Width is not provided. It will be set to the same value as the final dimension of the input shape.", UserWarning)
            self._width = self._input_shape[-1]
        elif isinstance(self._width, (list,tuple)):
            assert len(self._width) == self._depth, "If width is a list or tuple, it must have length equal to the depth."
            
            
        self._hidden_activation = hparams.get("hidden_activation")
        self._hidden_activation_params = hparams.get("hidden_activation_params")
        self._include_output_layer = hparams.get("include_output_layer")
        self._output_activation = hparams.get("output_activation")
        self._output_activation_params = hparams.get("output_activation_params")
        self._norm_layer_type = hparams.get("norm_layer_type")
        self._norm_layer_position = hparams.get("norm_layer_position")
        self._norm_layer_params = hparams.get("norm_layer_params")
        self.batch_input_shape = (self._batch_size,) + tuple(self._input_shape)
        self.batch_output_shape = (self._batch_size,) + tuple(self._input_shape[:-1]) + (self._output_size,)
        if self._output_activation is not None:
            if isinstance(self._output_activation, str):
                if self._output_activation.lower()==self._output_activation:
                    self._output_activation_module = getattr(tf.keras.activations, self._output_activation)
                else:
                    self._output_activation_module = getattr(tf.keras.layers, self._output_activation)
            else:
                self._output_activation_module = self._output_activation
        else:
            self._output_activation_module = None
        
        # Generate arrays containing parameters of each Dense Block (Every block contains a linear, normalization, activation, and dropout layer).
        self._dense_width_vec = self._gen_hparam_vec_for_dense(self._width, 'width')
        self._dense_activation_vec = self._gen_hparam_vec_for_dense(self._hidden_activation, 'hidden_activation')
        self._dense_activation_params_vec = self._gen_hparam_vec_for_dense(self._hidden_activation_params, 'hidden_activation_params')
        self._dense_norm_layer_type_vec = self._gen_hparam_vec_for_dense(self._norm_layer_type, 'norm_layer_type')
        self._dense_norm_layer_params_vec = self._gen_hparam_vec_for_dense(self._norm_layer_params, 'norm_layer_params')
        self._dense_norm_layer_position_vec = self._gen_hparam_vec_for_dense(self._norm_layer_position, 'norm_layer_position')
        self._dense_dropout_vec = self._gen_hparam_vec_for_dense(self._dropout, 'dropout')
        
        # Initialize sequential model
        self.net = tf.keras.models.Sequential(name=self._model_name)
        
        # Construct the dense layers
        # in_size = self._input_shape
        for i in range(self._depth):
            # out_size = self._dense_width_vec[i]
            _kwargs = {
                'model':self.net,
                'output_size':self._dense_width_vec[i],
                'activation':self._dense_activation_vec[i],
                'activation_params':self._dense_activation_params_vec[i],
                'norm_layer_type':self._dense_norm_layer_type_vec[i],
                'norm_layer_position':self._dense_norm_layer_position_vec[i],
                'norm_layer_params':self._dense_norm_layer_params_vec[i],
                'dropout':self._dense_dropout_vec[i],
                'kernel_regularizer':(tf.keras.regularizers.L2(self._l2_reg) if self._l2_reg else None)
            }
            if i==0 and self._input_shape is not None:
                _kwargs.update({'input_shape':self._input_shape})
            add_dense_block(**_kwargs)
            # in_size = out_size
        
        # Output layer
        if self._include_output_layer:
            self.net.add(tf.keras.layers.Dense(self._output_size, kernel_regularizer=(tf.keras.regularizers.L2(self._l2_reg) if self._l2_reg else None)))
            if self._output_activation:
                if isinstance(self._output_activation, str):
                    if self._output_activation.lower() == self._output_activation:
                        self.net.add(tf.keras.layers.Activation(self._output_activation))
                    elif self._output_activation_params:
                        self.net.add(getattr(tf.keras.layers, self._output_activation)(**self._output_activation_params))
                    else:
                        self.net.add(getattr(tf.keras.layers, self._output_activation)())
                elif self._output_activation_params:
                    self.net.add(self._output_activation(**self._output_activation_params))
                else:
                    self.net.add(self._output_activation())

    def _gen_hparam_vec_for_dense(self, hparam, hparam_name, **kwargs):
        return generate_array_for_hparam(hparam, self._depth, hparam_name=hparam_name, count_if_not_list_name='depth', **kwargs)
        
    def call(self, x, **kwargs):
        return self.net(x, **kwargs)





if __name__ == '__main__':
    
    hparams = {
        # General and I/O parameters
        "model_name": "My Model",
        "input_shape": [28, 28, 3], #[10],
        "output_size": 24,
        # Architecture parameters
        "width": 32,
        "depth": 3,
        "hidden_activation": "LeakyReLU",
        "hidden_activation_params": {'alpha':0.1},
        "norm_layer_type":"BatchNormalization",
        "norm_layer_position": "before",
        "norm_layer_params": None,
        "dropout": 0.2,
        "include_output_layer": True,
        "output_activation": "Softmax",
        "output_activation_params": None,
        # Training procedure parameters
        "learning_rate": 0.001,
        "exponential_decay_rate": 0.9,
        "batch_size": 32,
        "epochs": 2,
        "early_stopping_patience_epochs": 2,
        "validation_data": [0.05,'trainset'],
        "l2_reg": 0.0001,
        "loss_function": "categorical_crossentropy",
        'optimizer': 'Adam',
        'optimizer_params': None,
        'metrics': ['accuracy'],
        'checkpoint_path': None,
        'early_stopping_monitor': 'loss',
        'early_stopping_mode': 'min',
        'early_stopping_value': 1.0e-6
    }
    test_keras_model_class(ANN, hparams, False)
    


import os
import pprint
import re

import numpy as np
from keras import backend as K
from keras.models import Sequential
from keras.layers import (
    Activation, BatchNormalization, Conv2D, Dense, Dropout,
    ELU, Embedding, Flatten, LocallyConnected1D, LSTM, MaxPooling2D,
)

from keras2cpp import export_model

np.set_printoptions(precision=25, threshold=np.nan)


os.makedirs('test', exist_ok=True)
for path_ in os.listdir('test'):
    os.remove('test/' + path_)

os.makedirs('models', exist_ok=True)
for path_ in os.listdir('models'):
    os.remove('models/' + path_)


def c_array(a):
    s = pprint.pformat(a.flatten())

    s = re.sub(r'[ \t\n]*', '', s)
    s = re.sub(r'[ \t]*,[ \t]*', ', ', s)
    s = re.sub(r'[ \t]*\][, \t]*', '} ', s)
    s = re.sub(r'[ \t]*\[[ \t]*', '{', s)
    s = s.replace('array(', '').replace(')', '')
    s = re.sub(r'[, \t]*dtype=float32', '', s)
    s = s.strip()

    shape = ''
    if a.shape:
        shape = repr(a.shape)
        shape = re.sub(r',*\)', '}', shape.replace('(', '{'))
    else:
        shape = '{1}'
    return shape, s


TEST_CASE = '''/* Autogenerated file, DO NOT EDIT */
#pragma once

#include "src/model.h"

namespace test {
    inline auto %(name)s() {
        printf("TEST %(name)s\\n");
    #pragma GCC diagnostic push
    #pragma GCC diagnostic ignored "-Wconversion"
        keras2cpp::Tensor in%(x_shape)s;
        in.data_ = %(x_data)s;
        keras2cpp::Tensor target%(y_shape)s;
        target.data_ = %(y_data)s;
    #pragma GCC diagnostic pop
        auto [model, load_time] = keras2cpp::timeit(keras2cpp::Model::load, "%(path)s");
        auto [output, apply_time] = keras2cpp::timeit(model, in);
        for (size_t i = 0; i < target.dims_[0]; ++i)
            kassert_eq(target(i), output(i), %(eps)s);
        return std::make_tuple(load_time, apply_time);
    }
}
'''


def output_testcase(model, test_x, test_y, name, eps):
    print(f'Processing {name}')
    model.compile(loss='mse', optimizer='adam')
    model.fit(test_x, test_y, epochs=1, verbose=False)
    predict_y = model.predict(test_x).astype('f')
    print(model.summary())

    path = os.path.abspath(f'models/{name}.model')
    export_model(model, path)

    with open(f'test/{name}.h', 'w') as f:
        x_shape, x_data = c_array(test_x[0])
        y_shape, y_data = c_array(predict_y[0])

        f.write(TEST_CASE % dict(name=name, path=path, eps=eps,
                                 x_shape=x_shape,
                                 x_data=x_data,
                                 y_shape=y_shape,
                                 y_data=y_data))


# Dense 1x1
test_x = np.arange(10)
test_y = test_x * 10 + 1
model = Sequential([
    Dense(1, input_dim=1)
])
output_testcase(model, test_x, test_y, 'dense_1x1', '1e-6')


# Dense 10x1
test_x = np.random.rand(10, 10).astype('f')
test_y = np.random.rand(10).astype('f')
model = Sequential([
    Dense(1, input_dim=10)
])
output_testcase(model, test_x, test_y, 'dense_10x1', '1e-6')


# Dense 2x2
test_x = np.random.rand(10, 2).astype('f')
test_y = np.random.rand(10).astype('f')
model = Sequential([
    Dense(2, input_dim=2),
    Dense(1)
])
output_testcase(model, test_x, test_y, 'dense_2x2', '1e-6')


# Dense 10x10
test_x = np.random.rand(10, 10).astype('f')
test_y = np.random.rand(10).astype('f')
model = Sequential([
    Dense(10, input_dim=10),
    Dense(1)
])
output_testcase(model, test_x, test_y, 'dense_10x10', '1e-6')


# Dense 10x10x10
test_x = np.random.rand(10, 10).astype('f')
test_y = np.random.rand(10, 10).astype('f')
model = Sequential([
    Dense(10, input_dim=10),
    Dense(10)
])
output_testcase(model, test_x, test_y, 'dense_10x10x10', '1e-6')


# Conv 2x2
test_x = np.random.rand(10, 2, 2, 1).astype('f')
test_y = np.random.rand(10, 1).astype('f')
model = Sequential([
    Conv2D(1, (2, 2), input_shape=(2, 2, 1)),
    Flatten(),
    Dense(1)
])
output_testcase(model, test_x, test_y, 'conv_2x2', '1e-6')


# Conv 3x3
test_x = np.random.rand(10, 3, 3, 1).astype('f').astype('f')
test_y = np.random.rand(10, 1).astype('f')
model = Sequential([
    Conv2D(1, (3, 3), input_shape=(3, 3, 1)),
    Flatten(),
    Dense(1)
])
output_testcase(model, test_x, test_y, 'conv_3x3', '1e-6')


# Conv 3x3x3
test_x = np.random.rand(10, 10, 10, 3).astype('f')
test_y = np.random.rand(10, 1).astype('f')
model = Sequential([
    Conv2D(3, (3, 3), input_shape=(10, 10, 3)),
    Flatten(),
    BatchNormalization(),
    Dense(1)
])
output_testcase(model, test_x, test_y, 'conv_3x3x3', '1e-6')


# Activation ELU
test_x = np.random.rand(1, 10).astype('f')
test_y = np.random.rand(1, 1).astype('f')
model = Sequential([
    Dense(10, input_dim=10),
    ELU(alpha=0.5),
    Dense(1, activation='elu')
])
output_testcase(model, test_x, test_y, 'elu_10', '1e-6')


# Activation relu
test_x = np.random.rand(1, 10).astype('f')
test_y = np.random.rand(1, 10).astype('f')
model = Sequential([
    Dense(10, input_dim=10),
    Activation('relu')
])
output_testcase(model, test_x, test_y, 'relu_10', '1e-6')


# Dense relu
test_x = np.random.rand(1, 10).astype('f')
test_y = np.random.rand(1, 10).astype('f')
model = Sequential([
    Dense(10, input_dim=10, activation='relu'),
    Dense(10, input_dim=10, activation='relu'),
    Dense(10, input_dim=10, activation='relu')
])
output_testcase(model, test_x, test_y, 'dense_relu_10', '1e-6')


# Dense relu
test_x = np.random.rand(1, 10).astype('f')
test_y = np.random.rand(1, 10).astype('f')
model = Sequential([
    Dense(10, input_dim=10, activation='tanh'),
    Dense(10, input_dim=10, activation='tanh'),
    Dense(10, input_dim=10, activation='tanh')
])
output_testcase(model, test_x, test_y, 'dense_tanh_10', '1e-6')


# Conv softplus
test_x = np.random.rand(10, 2, 2, 1).astype('f')
test_y = np.random.rand(10, 1).astype('f')
model = Sequential([
    Conv2D(1, (2, 2), input_shape=(2, 2, 1), activation='softplus'),
    Flatten(),
    Dense(1)
])
output_testcase(model, test_x, test_y, 'conv_softplus_2x2', '1e-6')


# Conv hardsigmoid
test_x = np.random.rand(10, 2, 2, 1).astype('f')
test_y = np.random.rand(10, 1).astype('f')
model = Sequential([
    Conv2D(1, (2, 2), input_shape=(2, 2, 1), activation='hard_sigmoid'),
    Flatten(),
    Dense(1)
])
output_testcase(model, test_x, test_y, 'conv_hard_sigmoid_2x2', '1e-6')


# Conv sigmoid
test_x = np.random.rand(10, 2, 2, 1).astype('f')
test_y = np.random.rand(10, 1).astype('f')
model = Sequential([
    Conv2D(1, (2, 2), input_shape=(2, 2, 1), activation='sigmoid'),
    Flatten(),
    Dense(1)
])
output_testcase(model, test_x, test_y, 'conv_sigmoid_2x2', '1e-6')


# Maxpooling2D 1x1
test_x = np.random.rand(10, 10, 10, 1).astype('f')
test_y = np.random.rand(10, 1).astype('f')
model = Sequential([
    MaxPooling2D(pool_size=(1, 1), input_shape=(10, 10, 1)),
    Flatten(),
    Dense(1)
])
output_testcase(model, test_x, test_y, 'maxpool2d_1x1', '1e-6')


# Maxpooling2D 2x2
test_x = np.random.rand(10, 10, 10, 1).astype('f')
test_y = np.random.rand(10, 1).astype('f')
model = Sequential([
    MaxPooling2D(pool_size=(2, 2), input_shape=(10, 10, 1)),
    Flatten(),
    Dense(1)
])
output_testcase(model, test_x, test_y, 'maxpool2d_2x2', '1e-6')


# Maxpooling2D 3x2x2
test_x = np.random.rand(10, 10, 10, 3).astype('f')
test_y = np.random.rand(10, 1).astype('f')
model = Sequential([
    MaxPooling2D(pool_size=(2, 2), input_shape=(10, 10, 3)),
    Flatten(),
    Dense(1)
])
output_testcase(model, test_x, test_y, 'maxpool2d_3x2x2', '1e-6')


# Maxpooling2D 3x3x3
test_x = np.random.rand(10, 10, 10, 3).astype('f')
test_y = np.random.rand(10, 1).astype('f')
model = Sequential([
    MaxPooling2D(pool_size=(3, 3), input_shape=(10, 10, 3)),
    Flatten(),
    Dense(1)
])
output_testcase(model, test_x, test_y, 'maxpool2d_3x3x3', '1e-6')


# LSTM simple 7x20
test_x = np.random.rand(10, 7, 20).astype('f')
test_y = np.random.rand(10, 3).astype('f')
model = Sequential([
    LSTM(3, return_sequences=False, input_shape=(7, 20))
])
output_testcase(model, test_x, test_y, 'lstm_simple_7x20', '1e-6')


# LSTM simple stacked 16x9
test_x = np.random.rand(10, 16, 9).astype('f')
test_y = np.random.rand(10, 1).astype('f')
model = Sequential([
    LSTM(16, return_sequences=False, input_shape=(16, 9)),
    Dense(3, input_dim=16, activation='tanh'),
    Dense(1)
])
output_testcase(model, test_x, test_y, 'lstm_simple_stacked_16x9', '1e-6')


# LSTM stacked 64x83
test_x = np.random.rand(10, 64, 83).astype('f')
test_y = np.random.rand(10, 1).astype('f')
model = Sequential([
    LSTM(16, return_sequences=True, input_shape=(64, 83)),
    LSTM(16, return_sequences=False),
    Dense(1, activation='sigmoid')
])
output_testcase(model, test_x, test_y, 'lstm_stacked_64x83', '1e-6')


# Embedding 64
np.random.seed(10)
test_x = np.random.randint(100, size=(32, 10)).astype('f')
test_y = np.random.rand(32, 20).astype('f')
model = Sequential([
    Embedding(100, 64, input_length=10),
    Flatten(),
    # Dropout(0.5),
    Dense(20, activation='sigmoid')
])
output_testcase(model, test_x, test_y, 'embedding_64', '1e-6')


# Benchmark
test_x = np.random.rand(1, 128, 128, 3).astype('f')
test_y = np.random.rand(1, 10).astype('f')
model = Sequential([
    Conv2D(16, (7, 7), input_shape=(128, 128, 3), activation='relu'),
    MaxPooling2D(pool_size=(3, 3)),
    ELU(),
    Conv2D(8, (3, 3)),
    Flatten(),
    Dense(1000, activation='relu'),
    Dense(10)
])
output_testcase(model, test_x, test_y, 'benchmark', '1e-3')


os.system('clang-format -i --style=file test/*.h')



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

VERSION = '1.0.2' 
DESCRIPTION = "Easily build Keras models: utils for training/testing, built-in ANN, CNN, RNN models, modular Dense and Convolutional blocks, etc."

# Setting up
setup(
        # the name must match the folder name
        name="eznet_keras", 
        version=VERSION,
        author="Pouya P. Niaz",
        author_email="<pniaz20@ku.edu.tr>",
        url='https://github.com/pniaz20/eznet_keras',
        description=DESCRIPTION,
        long_description=long_description,
        long_description_content_type="text/markdown",
        packages=find_packages(),
        # packages=find_packages('eznet_keras'),
        # package_dir={'': 'eznet_keras'},
        python_requires=">=3.7, <4",
        license='MIT',
        install_requires=[
            'numpy','matplotlib','tensorflow(>=2.4)'
        ],
        keywords=['tensorflow','keras','deep learning','neural network','keras2cpp'],
        classifiers= [
            "Development Status :: 3 - Alpha",
            "Intended Audience :: End Users/Desktop",
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
            "Topic :: Utilities",
            "Topic :: Education",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Software Development :: Embedded Systems"
        ]
)


import struct
from functools import singledispatch

import numpy as np
from keras.layers import (
    Dense,
    Conv1D, Conv2D,
    LocallyConnected1D, LocallyConnected2D,
    Flatten,
    ELU,
    Activation,
    MaxPooling2D,
    LSTM,
    Embedding,
    BatchNormalization,
    )

LAYERS = (
    Dense,
    Conv1D, Conv2D,
    LocallyConnected1D, LocallyConnected2D,
    Flatten,
    ELU,
    Activation,
    MaxPooling2D,
    LSTM,
    Embedding,
    BatchNormalization,
)

ACTIVATIONS = (
    'linear',
    'relu',
    'elu',
    'softplus',
    'softsign',
    'sigmoid',
    'tanh',
    'hard_sigmoid',
    'softmax',
)


def write_tensor(f, data, dims=1):
    """
    Writes tensor as flat array of floats to file in 1024 chunks,
    prevents memory explosion writing very large arrays to disk
    when calling struct.pack().
    """
    for stride in data.shape[:dims]:
        f.write(struct.pack('I', stride))

    data = data.flatten()
    step = 1024
    written = 0

    for i in np.arange(0, len(data), step):
        remaining = min(len(data) - i, step)
        written += remaining
        f.write(struct.pack(f'={remaining}f', *data[i: i + remaining]))

    assert written == len(data)


def export_activation(activation, f):
    try:
        f.write(struct.pack('I', ACTIVATIONS.index(activation) + 1))
    except ValueError as exc:
        raise NotImplementedError(activation) from exc


@singledispatch
def export(layer, _):
    raise NotImplementedError(layer)


@export.register(Flatten)
def _(_0, _1):
    pass


@export.register(Activation)
def _(layer, f):
    activation = layer.get_config()['activation']
    export_activation(activation, f)


@export.register(ELU)
def _(layer, f):
    f.write(struct.pack('f', layer.alpha))


@export.register(BatchNormalization)
def _(layer, f):
    epsilon = layer.epsilon
    gamma = layer.get_weights()[0]
    beta = layer.get_weights()[1]
    pop_mean = layer.get_weights()[2]
    pop_variance = layer.get_weights()[3]

    weights = gamma / np.sqrt(pop_variance + epsilon)
    biases = beta - pop_mean * weights

    write_tensor(f, weights)
    write_tensor(f, biases)


@export.register(Dense)
def _(layer, f):
    # shape: (outputs, dims)
    weights = layer.get_weights()[0].transpose()
    biases = layer.get_weights()[1]
    activation = layer.get_config()['activation']

    write_tensor(f, weights, 2)
    write_tensor(f, biases)
    export_activation(activation, f)


@export.register(Conv1D)
def _(layer, f):
    # shape: (outputs, steps, dims)
    weights = layer.get_weights()[0].transpose(2, 0, 1)
    biases = layer.get_weights()[1]
    activation = layer.get_config()['activation']

    write_tensor(f, weights, 3)
    write_tensor(f, biases)
    export_activation(activation, f)


@export.register(Conv2D)
def _(layer, f):
    # shape: (outputs, rows, cols, depth)
    weights = layer.get_weights()[0].transpose(3, 0, 1, 2)
    biases = layer.get_weights()[1]
    activation = layer.get_config()['activation']

    write_tensor(f, weights, 4)
    write_tensor(f, biases)
    export_activation(activation, f)


@export.register(LocallyConnected1D)
def _(layer, f):
    # shape: (new_steps, outputs, ksize*dims)
    weights = layer.get_weights()[0].transpose(0, 2, 1)
    biases = layer.get_weights()[1]
    activation = layer.get_config()['activation']

    write_tensor(f, weights, 3)
    write_tensor(f, biases, 2)
    export_activation(activation, f)


@export.register(LocallyConnected2D)
def _(layer, f):
    # shape: (rows*cols, outputs, ksize*depth)
    weights = layer.get_weights()[0]
    # weights = weights.transpose(0, 2, 1)
    biases = layer.get_weights()[1]
    activation = layer.get_config()['activation']

    write_tensor(f, weights, 3)
    write_tensor(f, biases, 2)
    export_activation(activation, f)


@export.register(MaxPooling2D)
def _(layer, f):
    pool_size = layer.get_config()['pool_size']

    f.write(struct.pack('I', pool_size[0]))
    f.write(struct.pack('I', pool_size[1]))


@export.register(LSTM)
def _(layer, f):
    inner_activation = layer.get_config()['recurrent_activation']
    activation = layer.get_config()['activation']
    return_sequences = int(layer.get_config()['return_sequences'])

    weights = layer.get_weights()
    units = layer.units

    kernel, rkernel, bias = ([x[i: i+units] for i in range(0, 4*units, units)]
                             for x in (weights[0].transpose(),
                                       weights[1].transpose(),
                                       weights[2]))
    bias = [x.reshape(1, -1) for x in bias]
    for tensors in zip(kernel, rkernel, bias):
        for tensor in tensors:
            write_tensor(f, tensor, 2)

    export_activation(inner_activation, f)
    export_activation(activation, f)
    f.write(struct.pack('I', return_sequences))


@export.register(Embedding)
def _(layer, f):
    weights = layer.get_weights()[0]
    write_tensor(f, weights, 2)


def export_model(model, filename):
    with open(filename, 'wb') as f:
        layers = [layer for layer in model.layers
                  if type(layer).__name__ not in ['Dropout']]
        f.write(struct.pack('I', len(layers)))

        for layer in layers:
            f.write(struct.pack('I', LAYERS.index(type(layer)) + 1))
            export(layer, f)
